const NOP = () => {};

const forEach = (o, handler) => Object.keys(o).forEach(k => handler(o[k], k));

const __typePattern = /^\[object ([^\]]+)]$/;

const __getTypeOfObject = o => (__typePattern.exec(Object.prototype.toString.call(o)) || [undefined])[1];

const __serializeProp = (key, value, custom, serialized) => typeof value === 'string' ? `${key}="${value}"` : `${key}={${__serialize(value, custom, serialized)}}`;

const __serializeProps = (o, custom, serialized) => {
  const elemProps = o.props;
  const props = [];
  forEach(elemProps, (v, k) => {
    if (v !== undefined && k !== 'children') props.push(__serializeProp(k, v, custom, serialized));
  });
  if (o.key) props.push(__serializeProp('key', o.key, custom, serialized));
  if (o.ref) props.push(__serializeProp('ref', o.ref, custom, serialized));
  if (!props.length) return '';
  return ' ' + props.join(' ');
};

const __serializeChildren = (o, custom, serialized) => {
  const children = o.props.children;
  if (!children) return '';
  if (typeof children === 'string') return children;
  return Array.isArray(children) ? children.map(v => __serialize(v, custom, serialized)).join('') : __serialize(children, custom, serialized);
};

const __reactTypeNameReader = {
  String: type => type,
  Function: type => type.name,
  Symbol: type => Symbol.keyFor(type) === 'react.fragment' && 'Fragment' || undefined
};

const getTypeName = type => (__reactTypeNameReader[__getTypeOfObject(type)] || NOP)(type);

const __serializeReactElement = (o, custom, serialized) => {
  const type = getTypeName(o.type) || 'UNKNOWN';

  const children = __serializeChildren(o, custom, serialized);

  return `<${type}${__serializeProps(o, custom, serialized)}${children ? `>${children}</${type}>` : ' />'}`;
};

const __serializeReact = (o, custom, serialized) => {
  const symbolKey = Symbol.keyFor(o.$$typeof);

  switch (symbolKey) {
    case 'react.element':
      return __serializeReactElement(o, custom, serialized);

    default:
      return;
  }
};

const __serializeIfReact = (o, custom, serialized) => {
  if (Object.prototype.toString.call(o.$$typeof) === '[object Symbol]') {
    const key = Symbol.keyFor(o.$$typeof);
    if (key && key.indexOf('react.') === 0) return __serializeReact(o, custom, serialized);
  }
};

const __serializerByType = {
  BigInt: o => `${String(o)}n`,
  RegExp: o => `/${String(o)}/`,
  String: o => o.indexOf('"') === -1 && o.indexOf("'") !== -1 ? `"${o}"` : `'${o}'`,
  Function: o => o.name || 'Function',
  AsyncFunction: o => o.name || 'AsyncFunction',
  Date: o => `new Date(${Number(o)})`,
  Number: o => String(o),
  Boolean: o => String(o),
  Set: o => `new Set(${__serializeArray([...o.keys()], undefined, [])})`,
  Map: o => `new Map(${__serializeArray([...o.entries()], undefined, [])})`,
  Symbol: o => Symbol.keyFor(o) === undefined ? o.toString() : `Symbol.for('${Symbol.keyFor(o)}')`,
  Error: o => `new ${o.name}('${o.message}')`
};

const __serializeArray = (o, custom, serialized) => {
  const results = [];

  for (let i = 0; i < o.length; i++) {
    results.push(__serialize(o[i], custom, serialized));
  }

  return `[${results.join(', ')}]`;
};

const __serializeOptArray = (o, custom, serialized) => {
  if (Array.isArray(o)) return __serializeArray(o, custom, serialized);
};

const __serializeObject = (o, custom, serialized) => {
  const oKeys = Object.keys(o);
  oKeys.sort();
  const results = [];

  for (let i = 0; i < oKeys.length; i++) {
    const key = oKeys[i];
    results.push(`${key}: ${__serialize(o[key], custom, serialized)}`);
  }

  const objectType = o.constructor.name;
  const displayedType = objectType === 'Object' ? '' : objectType;
  return `${displayedType}{${results.join(', ')}}`;
};

const Serialize = {
  custom: (o, custom) => custom && custom(o),
  undefOrNull: o => o === undefined && 'undefined' || o === null && 'null' || undefined,
  flat: o => (__serializerByType[__getTypeOfObject(o)] || NOP)(o),
  cyclomatic: (o, _, serialized) => serialized.indexOf(o) !== -1 && '>CYCLOMATIC<' || undefined,
  react: __serializeIfReact,
  array: __serializeOptArray
};

const __serialize = (o, custom, serialized) => Serialize.custom(o, custom, serialized) || Serialize.undefOrNull(o, custom, serialized) || Serialize.flat(o, custom, serialized) || Serialize.cyclomatic(o, custom, serialized) || (nextSerialized => Serialize.react(o, custom, nextSerialized) || Serialize.array(o, custom, nextSerialized) || __serializeObject(o, custom, nextSerialized))([...serialized, o]);

const _serialize = (o, custom) => __serialize(o, custom, []);

const Serializer = {
  run: o => _serialize(o),
  create: custom => o => _serialize(o, custom)
};

/**
 * This file is part of spy4js which is released under MIT license.
 *
 * The LICENSE file can be found in the root directory of this project.
 *
 */
/**
 * This symbol serves as replacement to ignore any
 * inequality and skip further comparisons.
 */
const IGNORE = Symbol.for('__Spy_IGNORE__');
const serialize = Serializer.create((o) => (o === IGNORE && '>IGNORED<') || undefined);

/**
 * This file is part of spy4js which is released under MIT license.
 *
 * The LICENSE file can be found in the root directory of this project.
 *
 */
const SPY_COMPARE_FAILED = 'Spy.COMPARE failed';
const SPY_MAPPER_FAILED = 'Spy.MAPPER failed';
/**
 * Uniquely identifiable container for spy relevant comparators.
 */
class SpyComparator {
    constructor(comparator) {
        this._func = comparator;
    }
    compare(arg) {
        const result = this._func(arg);
        if (typeof result === 'string')
            return [`${SPY_MAPPER_FAILED} [${result}]`];
        if (result === false)
            return [SPY_COMPARE_FAILED];
        return undefined;
    }
}
/**
 * This function may create individual comparators
 * to define argument based comparison for arbitrary
 * nested objects.
 *
 * Returning "any" because this return value should be usable anywhere.
 */
const COMPARE = (comparator) => new SpyComparator(comparator);
/**
 * Returning "any" because this return value should be usable anywhere.
 */
const MAPPER = (from, to) => new SpyComparator((mapper) => {
    const result = mapper(...(Array.isArray(from) ? from : [from]));
    const diff = differenceOf(result, to);
    return diff ? `${serialize(result)} did not match ${serialize(to)}: ${diff}` : undefined;
});
const __different = (type) => ['different ' + type];
/**
 * This function is the internal representation of
 * "differenceOf". It does recursively call itself.
 * Read more below.
 *
 * @param {any} a <- any param.
 * @param {any} b <- any param to compare with the first param.
 * @param {boolean} useOwnEquals <-  enables/disables the usage
 *                                   of own "equals" implementations.
 * @param {Array<any>} alreadyComparedArray
 *      <- All flatly compared objects
 *         will be temporarily stored
 *         in this array to resolve
 *         circular structures.
 *
 * @return {string|void} <- information about the difference
 *                           of the provided arguments.
 */
const __diff = (a, b, useOwnEquals, alreadyComparedArray = []) => {
    if (a === IGNORE || b === IGNORE)
        return undefined;
    if (a instanceof SpyComparator)
        return a.compare(b);
    if (b instanceof SpyComparator)
        return b.compare(a);
    if (a === b)
        return undefined;
    if (a === undefined || b === undefined)
        return ['one was undefined'];
    if (a === null || b === null)
        return ['one was null'];
    const aClass = Object.prototype.toString.call(a);
    const bClass = Object.prototype.toString.call(b);
    if (aClass !== bClass)
        return __different(`object types: ${aClass} <-> ${bClass}`);
    switch (aClass) {
        case '[object RegExp]':
            if (String(a) === String(b)) {
                return undefined;
            }
            return __different('regexp');
        case '[object String]':
            return __different('string');
        case '[object Function]':
            return __different('function');
        case '[object AsyncFunction]':
            return __different('async function');
        case '[object Number]':
            if (isNaN(a) && isNaN(b)) {
                return undefined;
            }
            return __different('number');
        case '[object BigInt]':
            return __different('BigInt');
        case '[object Date]':
            if (Number(a) === Number(b)) {
                return undefined;
            }
            return __different('date');
        case '[object Boolean]':
            return __different('bool');
        case '[object Symbol]':
            return __different('symbols');
        case '[object Error]':
            if (String(a) === String(b)) {
                return undefined;
            }
            return __different('error');
        default:
            if (a.constructor !== b.constructor) {
                return __different('constructor');
            }
    }
    if (useOwnEquals && a.equals instanceof Function) {
        if (a.equals(b)) {
            return undefined;
        }
        return [
            'own equals method failed <- ' +
                'Maybe you want to disable the usage ' +
                'of own equals implementation? ' +
                '[ Use: spy.configure({useOwnEquals: false}) ]',
        ];
    }
    if (alreadyComparedArray.indexOf(a) !== -1) {
        return undefined;
    }
    const compared = [...alreadyComparedArray, a];
    const keys = Array.from(new Set(Object.keys(a).concat(Object.keys(b))));
    for (const key of keys) {
        const diff = __diff(a[key], b[key], useOwnEquals, compared);
        if (diff !== undefined) {
            return [key, ...diff];
        }
    }
    return undefined;
};
const __serializeDifferentProp = (obj, diff) => {
    // asserting diff.length > 1
    let toSerialize = obj;
    diff.slice(0, -1).forEach((key) => {
        toSerialize = toSerialize[key];
    });
    return serialize(toSerialize);
};
const __diffToStr = (diff, info = '') => {
    if (diff.length === 1)
        return diff[0];
    return `--> ${diff[0]} / ` + diff.slice(1).join(' / ') + info;
};
/**
 * This function does make a comparison of two provided params.
 *
 * If found any difference it will return a string containing
 * information about the first detected difference.
 *
 * If the given params were proven "identical" this function
 * returns undefined.
 *
 * In first place the params are flatly checked. The checks are
 * in the following order:
 *
 * - identical objects
 * - one param is null or undefined
 * - object type
 * - regexp/string/number/date/boolean
 * - constructor
 * - keys length
 * - own "equals" implementation
 *
 * In second place all keys of the params are deeply checked.
 * Using the above flat checks for all attributes.
 *
 * Also all circular structures will be resolved and repeating
 * attributes will be assumed to be equal.
 *
 * @param {any} a <- any param.
 * @param {any} b <- any param to compare with the first param.
 * @param {{useOwnEquals:boolean}} config <- controls the usage of own
 *                                           "equals" implementations.
 *
 * @return {string|void} <- information about the difference
 *                           of the provided arguments.
 */
const differenceOf = (a, b, config = { useOwnEquals: true }) => {
    const diff = __diff(a, b, config.useOwnEquals);
    if (!diff)
        return undefined;
    const diffStr = __diffToStr(diff);
    if (diff.length < 2)
        return diffStr;
    const diffProp1 = __serializeDifferentProp(a, diff);
    const lastPart = diff[diff.length - 1];
    if (lastPart.indexOf(SPY_MAPPER_FAILED) === 0)
        return diffStr;
    const info = lastPart === SPY_COMPARE_FAILED
        ? `called with: ${diffProp1}`
        : `${diffProp1} != ${__serializeDifferentProp(b, diff)}`;
    return `${diffStr} [${info}]`;
};
const toError = (msgOrError, spyName) => msgOrError instanceof Error ? msgOrError : new Error(msgOrError || `${spyName} was requested to throw`);

/**
 * This file is part of spy4js which is released under MIT license.
 *
 * The LICENSE file can be found in the root directory of this project.
 *
 */
const restoreAttributeForEntry = (value) => {
    const { obj, method, methodName } = value;
    if (obj) {
        obj[methodName] = method;
    }
};
class SpyRegistry {
    /**
     * The SpyRegistry is a class to handle the
     * correct restoration of spied objects.
     *
     * You may push an objects information about
     * a spied method to store it and be able to
     * restore it at any time.
     * Consider the SpyRegistry as information storage
     * for spied objects.
     *
     * @constructor
     */
    constructor() {
        this.register = {};
        this.persReg = {};
        this.registerCount = 0;
    }
    /**
     * If called, the SypRegistry will be resetting to its initial state.
     * Exception: the unique register count will not be touched.
     * Meaning that all stored object information will be restored
     * to their individual previous state.
     */
    restoreAll() {
        Object.values(this.register).forEach((entry) => {
            restoreAttributeForEntry(entry);
        });
        this.register = {};
    }
    /**
     * If called, the SpyRegistry will restore the object,
     * which was registered at given index and is getting lose
     * of the stored information.
     *
     * If the registry entry for the given index does not exist,
     * nothing will happen.
     *
     * @param {number} index -> the unique identifier of stored information.
     */
    restore(index) {
        const entry = this.register[index];
        if (entry) {
            restoreAttributeForEntry(entry);
            delete this.register[index];
        }
    }
    /**
     * If called, the SpyRegistry will store the given information.
     * The unique identifier index will be returned.
     *
     * @param {Object} obj -> The related object, which will be spied.
     * @param {string} methodName -> The name of the mocked method.
     * @return {number} -> The unique store index.
     */
    push(obj, methodName) {
        this.registerCount += 1;
        this.register[this.registerCount] = {
            obj,
            method: obj[methodName],
            methodName,
        };
        return this.registerCount;
    }
    /**
     * If called, the stored method for the corresponding index
     * will be returned. If the registry entry does not exist,
     * undefined will be returned.
     *
     * @param {number} index -> the unique identifier of stored information.
     * @return {any} -> Any stored information can be returned.
     *                   BUT: Usually this method returns a function or
     *                        undefined.
     */
    getOriginalMethod(index) {
        const entry = this.register[index];
        if (entry) {
            return entry.method;
        }
    }
    /**
     * If called, the stored method will be moved from the standard
     * registry into the persistent registry or vice versa.
     * This can make restore and restoreAll having no effect anymore.
     *
     * @param {number} index -> the unique identifier of stored information.
     * @param {boolean} intoPersReg -> boolean to determine the moving
     *                                 direction.
     */
    persist(index, intoPersReg) {
        const fromReg = intoPersReg ? this.register : this.persReg;
        const toReg = intoPersReg ? this.persReg : this.register;
        const entry = fromReg[index];
        if (entry) {
            toReg[index] = entry;
            delete fromReg[index];
        }
    }
}

/**
 * This file is part of spy4js which is released under MIT license.
 *
 * The LICENSE file can be found in the root directory of this project.
 *
 */
const uninitialized = (method) => () => {
    throw new Error(`Method '${String(method)}' was not initialized on Mock.`);
};
const defaultScope = Symbol('__Spy_global__');
const _mocks = { [defaultScope]: [] };
let scope = defaultScope;
const setScope = (scoping) => {
    if (scoping) {
        _mocks[scoping] = [];
        scope = scoping;
    }
    else
        scope = defaultScope;
};
const registerMock = (mocked, returns) => {
    const mock = {};
    _mocks[scope].push({ mocked, mock, scope, returns });
    return mock;
};
const createMock = (obj, methods, returns) => {
    const mock = registerMock(obj, returns);
    methods.forEach((method) => {
        mock[method] = uninitialized(method);
    });
    return mock;
};
const couldNotInitError = (scope, additional) => new Error(`Could not initialize mock for ${scope === defaultScope ? 'global scope' : `scope "${scope}"`}, because:\n${additional}`);
const initMock = ({ mocked, mock, scope, returns }, spyOn) => {
    Object.keys(mock).forEach((method) => {
        try {
            mock[method] = spyOn(mocked, method).returns(returns);
        }
        catch (e) {
            throw couldNotInitError(scope, e.message);
        }
    });
};
const initMockScope = (scoping, spyOn) => {
    Object.values(_mocks[scoping]).forEach((mock) => initMock(mock, spyOn));
};
const initMocks = (spyOn, scoping) => {
    initMockScope(defaultScope, spyOn);
    scoping && initMockScope(scoping, spyOn);
};

/**
 * This file is part of spy4js which is released under MIT license.
 *
 * The LICENSE file can be found in the root directory of this project.
 *
 */
// 'path' and 'fs' have to be lazy loaded because we first need to validate
// if the CommonJS context is used by the test runner.
const pathDirname = (p) => require('path').dirname(p);
const pathJoin = (...p) => require('path').join(...p);
const fsExistsSync = (p) => require('fs').existsSync(p);
// we have to cheat here flow since it might be the case that "jest" was not
// configured correctly or the jest test runner might not even be present
// $FlowFixMe
const _testSuite = { isJest: !!jest, isCJS: !!require };
const runner = {};
const oldDescribe = describe;
// eslint-disable-next-line
const newDescribe = (name, suite) => {
    oldDescribe(name, () => {
        const scoping = String(name);
        setScope(scoping);
        beforeEach(() => {
            runner.beforeEach && runner.beforeEach(scoping);
        });
        afterEach(() => {
            runner.afterEach && runner.afterEach(scoping);
        });
        const rv = suite();
        setScope(undefined);
        return rv;
    });
};
newDescribe.each = oldDescribe.each;
newDescribe.only = oldDescribe.only;
newDescribe.skip = oldDescribe.skip;
// eslint-disable-next-line no-global-assign
describe = newDescribe;
const configure = (other) => {
    if (other.afterEach)
        runner.afterEach = other.afterEach;
    if (other.beforeEach)
        runner.beforeEach = other.beforeEach;
};
const addSnapshotSerializer = (serializer) => {
    _testSuite.isJest && expect && expect.addSnapshotSerializer && expect.addSnapshotSerializer(serializer);
};
// 1. Spy.createMock in some test
// 2. _createMock from test-suite.js
// 3. __getAbsolutePath or __getNodeModulePath from test-suite.js
// 4. __callerBasedir from test-suite.js
// 5. __caller from test-suite.js
const STACK_NUM_CREATE_MOCK = 5;
const __caller = () => {
    const traceFn = Error.prepareStackTrace;
    Error.prepareStackTrace = (_error, stack) => stack;
    const stack = new Error().stack;
    Error.prepareStackTrace = traceFn;
    // it highly depends from were the functionality will be called
    return stack[STACK_NUM_CREATE_MOCK].getFileName();
};
const __callerBasedir = () => pathDirname(__caller());
const __getAbsolutePath = (moduleName) => pathJoin(__callerBasedir(), moduleName);
const __getDepDir = (moduleName, file) => (moduleName.includes('/') ? pathDirname(file) : file);
// We cannot just use "require(moduleName)" because we need to derive the correct
// "node_modules" to use for the caller.
// E.g. if "spy4js" is installed globally or in a workspace root while
//      the caller wants to mock a module contained several levels deeper in
//      the file tree, we need to determine the correct dependency like the Node
//      implementation of "require" does.
// Behavior: From the caller base dir iterate the file tree up until the desired
//           module was found in some "node_modules" directory. Fail if the
//           module could not be found in the parent hierarchy.
const __getNodeModulePath = (moduleName) => {
    let baseDir = __callerBasedir();
    const calcFile = () => pathJoin(baseDir, 'node_modules', moduleName);
    let file = calcFile();
    while (!fsExistsSync(__getDepDir(moduleName, file))) {
        if (baseDir === '/') {
            throw new Error(`spy4js: Could not find given module: "${moduleName}"`);
        }
        baseDir = pathJoin(baseDir, '..');
        file = calcFile();
    }
    return file;
};
const createModuleMock = (moduleName, names, returns) => {
    if (!_testSuite.isCJS)
        throw new Error('spy4js: Mocking a module only works if your test runner executes with CommonJS');
    const isNodeModule = moduleName.indexOf('.') !== 0 && moduleName.indexOf('/') !== 0;
    // now we are free to use "require('path')" to calculate the correct
    // module path for the mocking.
    return createMock(require(isNodeModule ? __getNodeModulePath(moduleName) : __getAbsolutePath(moduleName)), names, returns);
};
const TestSuite = { addSnapshotSerializer, createModuleMock, configure };

/**
 * This file is part of spy4js which is released under MIT license.
 *
 * The LICENSE file can be found in the root directory of this project.
 *
 */
const CallOrder = {
    Idx: 0,
    lastCheck: 0,
};
/**
 *
 * Instantiating the SpyRegistry to handle all
 * mocked object relative information and
 * restore their functionality if requested.
 *
 */
const registry = new SpyRegistry();
let __LOCK__ = true;
/**
 * Those symbols are used to protect the private spy properties from outer manipulation by mistake.
 */
const symbolName = Symbol('__Spy_name__');
const symbolSnap = Symbol('__Spy_snap__');
const symbolIsSpy = Symbol('__Spy_isSpy__');
const symbolFunc = Symbol('__Spy_func__');
const symbolCalls = Symbol('__Spy_calls__');
const symbolConfig = Symbol('__Spy_config__');
const symbolIndex = Symbol('__Spy_index__');
const Symbols = {
    name: symbolName,
    snap: symbolSnap,
    isSpy: symbolIsSpy,
    func: symbolFunc,
    calls: symbolCalls,
    config: symbolConfig,
    index: symbolIndex,
};
/**
 * Very jest specific snapshot serialization behavior.
 *
 * Hint: We are casting here anything to any, because not all users might
 * have typed those functions correctly and should not see flow errors
 * related to those types.
 */
TestSuite.addSnapshotSerializer({
    test: (v) => v && v[Symbols.isSpy],
    print: (spy) => spy[Symbols.snap],
});
/**
 * Initial default settings for every
 * spy instance. Can be modified only
 * implicitly by "Spy.configure".
 *
 * @type {{useOwnEquals: boolean}}
 */
const DefaultSettings = {
    useOwnEquals: true,
    enforceOrder: false,
};
const SpyHelperFunctions = {
    getCalls: (spy) => {
        const allCalls = spy[Symbols.calls];
        if (!DefaultSettings.enforceOrder)
            return allCalls;
        return allCalls.filter(({ order }) => order > CallOrder.lastCheck);
    },
};
const SpyFunctions = {
    /**
     * Configures this spy behavior in a special
     * way. Passing in an object that contains
     * meaningful attributes can configure:
     *
     * - useOwnEquals:boolean -> toggles the usage of own
     *                           implementation of "equals"
     *                           matcher, e.g. for comparing
     *                           call params with "wasCalledWith".
     *
     * - persistent:boolean -> toggles the persistence of the spy.
     *                         I.e. making it restorable or not.
     *                         Throws for not mocking spies.
     *
     * @param {Object} config <- An object containing attributes
     *                         for special configuration
     * @return {SpyInstance} <- BuilderPattern.
     */
    configure(config) {
        if (config.useOwnEquals !== undefined) {
            this[Symbols.config].useOwnEquals = config.useOwnEquals;
        }
        if (config.persistent !== undefined) {
            if (!this[Symbols.index]) {
                throw new Error(`\n\n${this[Symbols.name]} can not` +
                    ' be configured to be persistent!' +
                    ' It does not mock any object.');
            }
            this[Symbols.config].persistent = config.persistent;
            registry.persist(this[Symbols.index], this[Symbols.config].persistent);
        }
        return this;
    },
    /**
     * Accepts multiple functions. If called more often,
     * calling always the last supplied function.
     *
     * @param {Array<Function>} funcs
     *     -> The iterative provided functions
     *        can be accessed as array.
     *        And will be called one by one
     *        for each made call on the spy.
     *
     * @return {SpyInstance} <- BuilderPattern.
     */
    calls(...funcs) {
        if (funcs.length === 0) {
            // no arguments provided
            this[Symbols.func] = () => { };
            return this;
        }
        const max = funcs.length - 1;
        let counter = -1;
        this[Symbols.func] = (...args) => {
            counter++;
            return funcs[max < counter ? max : counter](...args);
        };
        return this;
    },
    /**
     * Accepts multiple return values. If called more often,
     * returns always the last supplied return value.
     *
     * @param {Array<any>} args -> The iterative provided arguments
     *                             can be accessed as array.
     *                             And will be returned one by one
     *                             for each made call.
     *
     * @return {SpyInstance} <- BuilderPattern.
     */
    returns(...args) {
        return this.calls(...args.map((arg) => () => arg));
    },
    /**
     * Accepts multiple values, which will be resolved sequentially.
     * If called more often, resolves always the last supplied value.
     *
     * @param {Array<any>} args -> The iterative provided arguments
     *                             can be accessed as array.
     *                             And will be resolved one by one
     *                             for each made call.
     *
     * @return {SpyInstance} <- BuilderPattern.
     */
    resolves(...args) {
        return this.returns(...(args.length ? args : [undefined]).map((arg) => Promise.resolve(arg)));
    },
    /**
     * Accepts multiple values, which will be rejected sequentially.
     * If called more often, rejects always the last supplied value.
     *
     * @param {Array<OptionalMessageOrError>} msgOrErrors
     *              -> The iterative provided arguments
     *                 can be accessed as array.
     *                 And will be rejected one by one
     *                 for each made call.
     *
     * @return {SpyInstance} <- BuilderPattern.
     */
    rejects(...msgOrErrors) {
        return this.calls(...(msgOrErrors.length ? msgOrErrors : [undefined]).map((msgOrError) => () => Promise.reject(toError(msgOrError, this[Symbols.name]))));
    },
    /**
     * Will make the spy throw an Error, if called next time.
     * The error message can be provided as parameter.
     *
     * @param {OptionalMessageOrError} msgOrError -> Will be the error message.
     *
     * @return {SpyInstance} <- BuilderPattern
     */
    throws(msgOrError) {
        this[Symbols.func] = () => {
            throw toError(msgOrError, this[Symbols.name]);
        };
        return this;
    },
    /**
     * Deletes all notices of made calls with this spy.
     *
     * @return {SpyInstance} <- BuilderPattern
     */
    reset() {
        this[Symbols.calls] = [];
        return this;
    },
    /**
     * Restores the last by this spy manipulated object
     * and removes this special mock.
     *
     * Restoring objects does not disable any
     * other behaviors/features of the spies.
     *
     * If the spy was configured persistent, than this
     * method will throw an exception.
     *
     * Other than "Spy.restoreAll" this method only removes
     * a maximum of one mock.
     *
     * @return {SpyInstance} <- BuilderPattern
     */
    restore() {
        if (this[Symbols.config].persistent) {
            throw new Error(`\n\n${this[Symbols.name]} can not be restored!` + ' It was configured to be persistent.');
        }
        registry.restore(this[Symbols.index]);
        return this;
    },
    /**
     * Makes the spy behave like the mocked
     * function. If no function was mocked by
     * this spy, it will do nothing if called.
     *
     * This function works exactly like
     * spy.transparentAfter(0).
     *
     * For example:
     * const spy = Spy.on(someObject, 'someFunc');
     * someObject.someFunc(); // calls only the spy
     * spy.transparent();
     * someObject.someFunc(); // behaves like calling the original method
     *
     * @return {SpyInstance} <- BuilderPattern
     */
    transparent() {
        return this.transparentAfter(0);
    },
    /**
     * If called with n as callCount this will make
     * the spy call the mocked function after called
     * the n'th time. For any spy that does not mock
     * any objects attribute, this will make the spy
     * do nothing if called after the n'th time.
     *
     * If the mocked function will get called again,
     * the made calls will still be registered.
     *
     * For example:
     * Spy.on(someObject, 'someFunc').transparentAfter(2);
     * someObject.someFunc(); // calls only the spy
     * someObject.someFunc(); // calls only the spy
     * someObject.someFunc(); // behaves like calling the original method
     *
     * @param {number} callCount <- The number after which the mocked function
     *                              should be called again.
     *
     * @return {SpyInstance} <- BuilderPattern
     */
    transparentAfter(callCount) {
        const oldFunc = this[Symbols.func];
        this[Symbols.func] = (...args) => {
            // before the function call is executed,
            // the call arguments were already saved
            // -> so we are interested if the made calls
            //    are more than the call count were we
            //    need to modify the behavior
            if (this[Symbols.calls].length > callCount) {
                const originalMethod = registry.getOriginalMethod(this[Symbols.index]);
                if (originalMethod) {
                    return originalMethod(...args);
                }
                return;
            }
            return oldFunc(...args);
        };
        return this;
    },
    /**
     * Checks if the spy was called callCount times often.
     *
     * If callCount is not provided then it only
     * checks if the spy was called at least once.
     *
     * Throws an error if the expectation is wrong.
     *
     * @param {?number} callCount -> Is the number of expected calls made.
     */
    wasCalled(callCount) {
        const calls = SpyHelperFunctions.getCalls(this);
        const madeCalls = calls.length;
        if (callCount !== undefined) {
            // check exactly the number of calls
            if (madeCalls !== callCount) {
                throw new Error(`\n\n${this[Symbols.name]} was called ${madeCalls} times,` +
                    ` but there were expected ${callCount} calls.\n\n` +
                    'Actually there were:\n\n' +
                    this.showCallArguments());
            }
        }
        else if (madeCalls === 0) {
            if (this[Symbols.calls].length) {
                throw new Error(`\n\n${this[Symbols.name]} was not called,` +
                    ` but there was expected at least one call.\n\n` +
                    'Actually there were:\n\n' +
                    this.showCallArguments());
            }
            else {
                throw new Error(`\n\n${this[Symbols.name]} was never called!\n\n`);
            }
        }
        if (callCount !== 0) {
            const checked = calls[(callCount || 1) - 1];
            CallOrder.lastCheck = checked.order;
        }
    },
    /**
     * Checks if the spy was call history matches the expectation.
     *
     * The call history has to match the call count and order.
     * Single arguments will be automatically wrapped as array, e.g.:
     *            1, 2, 3 -> [1], [2], [3]
     * ** Inspired by jest test.each **
     *
     * Throws an error if the expectation is wrong.
     *
     * @param {Array<Array<any> | any>} callHistory
     *          -> Are the expected made call arguments in correct order.
     */
    hasCallHistory(...callHistory) {
        const calls = SpyHelperFunctions.getCalls(this);
        const madeCalls = calls.length;
        const callCount = callHistory.length;
        if (madeCalls !== callCount) {
            throw new Error(`\n\n${this[Symbols.name]} was called ${madeCalls} times,` +
                ` but the expected call history includes exactly ${callHistory.length} calls.\n\n` +
                'Actually there were:\n\n' +
                this.showCallArguments());
        }
        if (callCount === 0) {
            throw new Error("\n\nPlease use 'wasNotCalled' instead of calling 'hasCallHistory' without any arguments");
        }
        const modifiedCallHistory = callHistory.map((arg) => (Array.isArray(arg) ? arg : [arg]));
        let hasErrors = false;
        const diffInfo = calls.map((call, index) => {
            const diff = differenceOf(call.args, modifiedCallHistory[index], this[Symbols.config]);
            if (diff)
                hasErrors = true;
            return diff;
        });
        if (hasErrors)
            throw new Error(`\n\n${this[Symbols.name]} was considered` +
                ' to be called with the following arguments in the given order:\n\n' +
                `${modifiedCallHistory
                    .map((entry, index) => `call ${index}: ${serialize(entry)}`)
                    .join('\n')}\n\n` +
                'Actually there were:\n\n' +
                this.showCallArguments(diffInfo));
        CallOrder.lastCheck = calls[callCount - 1].order;
    },
    /**
     * Checks that the spy was never called.
     * Throws an error if the spy was called at least once.
     */
    wasNotCalled() {
        const calls = SpyHelperFunctions.getCalls(this);
        const madeCalls = calls.length;
        if (madeCalls !== 0) {
            throw new Error(`\n\n${this[Symbols.name]} was not considered to be called, but was called ${madeCalls} times.\n\n` +
                'Actually there were:\n\n' +
                this.showCallArguments());
        }
    },
    /**
     * Checks if the spy was called with the provided arguments.
     *
     * Throws an error if the expectation is wrong.
     *
     * For example:
     * const spy = Spy();
     * spy(arg1, arg2, arg3);
     * spy(arg4, arg5);
     * spy.wasCalledWith(arg1, arg2, arg3); // no error
     * spy.wasCalledWith(arg4, arg5); // no error
     * spy.wasCalledWith(arg1); // error!!!
     *
     * @param {Array<any>} args -> The expected arguments
     *                           for any made call.
     */
    wasCalledWith(...args) {
        const calls = SpyHelperFunctions.getCalls(this);
        const madeCalls = calls.length;
        const diffInfo = [];
        for (let i = 0; i < madeCalls; i++) {
            const checked = calls[i];
            const diff = differenceOf(checked.args, args, this[Symbols.config]);
            if (!diff) {
                CallOrder.lastCheck = checked.order;
                return;
            }
            diffInfo.push(diff);
        }
        throw new Error(`\n\n${this[Symbols.name]} was considered` +
            ' to be called with the following arguments:\n\n' +
            `    --> ${serialize(args)}\n\n` +
            'Actually there were:\n\n' +
            this.showCallArguments(diffInfo));
    },
    /**
     * Checks if the spy was NOT called with the provided arguments.
     * This method checks the direct opposite of the method
     * spy.wasCalledWith.
     *
     * It throws an error if the upper method would not.
     *
     * For example:
     * const spy = Spy();
     * spy(arg1, arg2, arg3);
     * spy(arg4, arg5);
     * spy.wasCalledWith(arg1); // no error
     * spy.wasCalledWith(arg4, arg3); // no error
     * spy.wasCalledWith(arg4, arg5); // error!!!
     *
     * @param {Array<any>} args -> The not expected arguments
     *                             for any made call.
     */
    wasNotCalledWith(...args) {
        let errorOccurred = false;
        try {
            this.wasCalledWith(...args);
        }
        catch (e) {
            errorOccurred = true;
        }
        if (!errorOccurred) {
            throw new Error(`\n\n${this[Symbols.name]} was called` +
                ' unexpectedly with the following arguments:\n\n' +
                `    --> ${serialize(args)}\n\n`);
        }
    },
    /**
     * This method returns all call arguments of all calls. Will be an empty
     * array if the spy was never called.
     *
     * For example:
     * const spy = Spy();
     * spy(arg1, arg2, arg3);
     * spy(arg4, arg5, arg6);
     * spy.getAllCallArguments();
     * // returns [[arg1, arg2, arg3], [arg4, arg5, arg6]]
     *
     * @return {Array<any[]>} -> the call arguments of all calls.
     */
    getAllCallArguments() {
        return this[Symbols.calls].map(({ args }) => args);
    },
    /**
     * This method returns the call arguments of the
     * n'th made call as array. If less than n calls were made,
     * it will throw an error.
     *
     * By default n = 1. This corresponds to callNr = 0.
     *
     * For example:
     * const spy = Spy();
     * spy(arg1, arg2, arg3);
     * spy.getCallArguments(); // returns [arg1, arg2, arg3]
     *
     * @param {number} callNr -> represents the callNr for which
     *                           the call argument should be returned.
     *
     * @return {Array<any>} -> the call arguments of the (callNr + 1)'th call.
     */
    getCallArguments(callNr = 0) {
        const madeCalls = this.getAllCallArguments();
        if (callNr % 1 !== 0 || callNr >= madeCalls.length) {
            throw new Error(`\n\nThe provided callNr "${callNr}" was not valid.\n\n` +
                `Made calls for ${this[Symbols.name]}:\n\n` +
                this.showCallArguments());
        }
        return madeCalls[callNr];
    },
    /**
     * This method returns the m'th call argument of the
     * n'th made call. If less than n calls were made, it will throw
     * an error.
     *
     * By default n = 1. This corresponds to callNr = 0.
     * By default m = 1. This corresponds to argNr = 0.
     *
     * For example:
     * const spy = Spy();
     * spy(arg1, arg2, arg3);
     * spy(arg4, arg5, arg6);
     * spy.getCallArgument() === arg1; // true
     * spy.getCallArgument(1) === arg4; // true
     * spy.getCallArgument(0, 2) === arg3; // true
     * spy.getCallArgument(1, 1) === arg5; // true
     *
     * spy.getCallArgument(1, 5) === undefined; // true
     * spy.getCallArgument(2); // throws an exception
     *
     * @param {number} callNr -> represents the callNr for which
     *                           a call argument should be returned.
     * @param {number} argNr -> represents position of the argument
     *                          when the corresponding call was made.
     *
     * @return {any} -> the (argNr + 1)'th call argument
     *                  of the (callNr + 1)'th call.
     */
    getCallArgument(callNr = 0, argNr = 0) {
        return this.getCallArguments(callNr)[argNr];
    },
    /**
     * This method returns the number of made calls on the spy.
     *
     * @return {number} -> the number of made calls.
     */
    getCallCount() {
        return this.getAllCallArguments().length;
    },
    /**
     * This method returns a formatted text string for debugging
     * made calls with the given Spy. It is used also internally
     * if some wrong assertions were made on the Spy.
     * Some sample:
     *
     * call 0: [{"_key":"test1"}]
     * call 1: [{"_key":"test1"},{"_key":"test2"}]
     * call 2: [{"_key":"test3"},{"_key":"test2"},{"_key":"test1"}]
     * call 3: [{"_key":"test2"}]
     *
     * If an array of strings is provided, the given strings will
     * be printed just below params of each call.
     *
     * Some sample: additionalInformation = [
     *     '-> 0 / _key / different string',
     *     '-> 1 / _key / different object types'
     * ]
     *
     * call 0: [{"_key":"test1"}]
     *         -> 0 / _key / different string
     * call 1: [{"_key":"test1"},{"_key":"test2"}]
     *         -> 1 / _key / different object types
     * call 2: [{"_key":"test3"},{"_key":"test2"},{"_key":"test1"}]
     * call 3: [{"_key":"test2"}]
     *
     * @param {Array<string>} additionalInformation
     *      -> will be displayed below each call information
     *         as additional information.
     *
     * @return {string} -> The information about made calls.
     */
    showCallArguments(additionalInformation = []) {
        const allMadeCalls = this[Symbols.calls];
        const madeCalls = SpyHelperFunctions.getCalls(this);
        const earlierCalls = allMadeCalls.length - madeCalls.length;
        const usedAdditionalInfo = new Array(earlierCalls)
            .fill('!!! was called earlier (#enforceOrder)')
            .concat(additionalInformation);
        if (allMadeCalls.length === 0) {
            return `${this[Symbols.name]} was never called!\n`;
        }
        let response = '';
        for (let i = 0; i < allMadeCalls.length; i++) {
            const args = serialize(allMadeCalls[i].args);
            response += `call ${i - earlierCalls}: ${args}\n`;
            if (usedAdditionalInfo[i]) {
                response += `        ${usedAdditionalInfo[i]}\n`;
            }
        }
        return response;
    },
};
const AllCreatedSpies = [];
const _createSpy = (name = '', __mock) => {
    const spy = function (...args) {
        spy[Symbols.calls].push({ args, order: ++CallOrder.Idx });
        return spy[Symbols.func](...args);
    };
    if (__mock && !__LOCK__) {
        spy[Symbols.index] = registry.push(__mock.obj, __mock.methodName);
        spy[Symbols.name] = `the spy on '${name}'`;
        spy[Symbols.snap] = `Spy.on(${name})`;
    }
    else {
        spy[Symbols.index] = 0;
        spy[Symbols.name] = name || 'the spy';
        spy[Symbols.snap] = `Spy(${name})`;
    }
    spy[Symbols.isSpy] = true;
    spy[Symbols.func] = () => { };
    spy[Symbols.calls] = [];
    spy[Symbols.config] = { ...DefaultSettings };
    Object.entries(SpyFunctions).forEach(([key, value]) => {
        spy[key] = value;
    });
    AllCreatedSpies.push(spy);
    return spy;
};
/**
 * This constructor does instantiate a Spy
 * object.
 *
 * This spy is callable.
 * It does inherit all Spy specific methods below.
 * It holds additional (private) fields:
 * _name:string -> Will be displayed in all displayed
 *                 error messages.
 * _isSpy:boolean -> Always true for spies.
 * _func:Function -> The internal function, that will
 *                   actually we called, when calling
 *                   the spy.
 * _calls:Array<{arguments:Array<any>}>
 *     -> Stores the arguments with whom the spy was called.
 *        Each call adds another entry in the calls array.
 * _config = {useOwnEquals: boolean} -> internal spy config.
 *
 *
 * @param {string} name -> the identifier of the spy.
 *                       Useful for debugging issues.
 */
const Spy = (name) => _createSpy(name);
/**
 * This static method can be used to configure
 * the default behavior of created spy instances.
 * The most suited place where you could configure
 * spy4js is the "setupTests"-File, which runs
 * before each test suite.
 *
 * For example,
 *
 * Spy.configure({ useOwnEquals: false });
 *
 * would initially configure every spy to not
 * favor own "equals" implementation while
 * comparing any objects.
 *
 * You may also override default test suite hooks
 * by providing afterEach or beforeEach respectively.
 *
 * @param {Object} config <- Holds the configuration params.
 */
Spy.configure = (config) => {
    if (config.useOwnEquals !== undefined) {
        DefaultSettings.useOwnEquals = config.useOwnEquals;
    }
    if (config.enforceOrder !== undefined) {
        DefaultSettings.enforceOrder = config.enforceOrder;
    }
    TestSuite.configure({
        afterEach: config.afterEach,
        beforeEach: config.beforeEach,
    });
};
/**
 * This static attribute can be used to ignore the match
 * of a specific argument when using "wasCalledWith".
 */
Spy.IGNORE = IGNORE;
/**
 * This static attribute can be called with a custom
 * comparator that returns a boolean indicating if the
 * comparison holds. Can be used when calling e.g. "wasCalledWith".
 */
Spy.COMPARE = COMPARE;
/**
 * This static attribute can be used to test mapper
 * functions inside call argument checks. It is similar to
 * Spy.COMPARE, but wraps some mechanics to evaluate the result
 * based on the given input.
 */
Spy.MAPPER = MAPPER;
/**
 * This static method is an alternative way to
 * create a Spy which mocks the an objects attribute.
 *
 * The attribute of the object "obj[methodName]" will
 * be replaced by the spy and the previous attribute
 * will be stored in the spy registry.
 * Therefore this information is always restorable.
 * The most common use case, will be to mock
 * another function as attribute of the object.
 *
 * The method has to met the following conditions:
 *
 * - The attribute to spy has to be function itself.
 * - The attribute to spy should not be spied already.
 *
 * If the upper conditions are not fulfilled, this
 * method will throw to avoid unexpected behavior.
 *
 * @param {Object} obj -> The manipulated object.
 * @param {string} methodName -> The mocked attributes name.
 *
 * @return {SpyInstance}
 */
Spy.on = (obj, methodName) => {
    const method = obj[methodName];
    if (!(typeof method === 'function')) {
        throw new Error(`The object attribute '${methodName}' ` +
            `was: ${serialize(method)}\n\n` +
            'You should only spy on functions!');
    }
    if (method[Symbols.isSpy]) {
        throw new Error(`The objects attribute '${methodName}'` +
            ' was already spied. Please make sure to spy' +
            ' only once at a time at any attribute.');
    }
    __LOCK__ = false;
    const spy = _createSpy(methodName, { obj, methodName });
    __LOCK__ = true;
    obj[methodName] = spy;
    return spy;
};
/**
 * This static method is not only a shortcut for applying
 * multiple spies on one object at (different) attributes,
 * but it enables more control, clarity and comfort for all
 * kind of unit tests. (see spy.mock.test.js)
 *
 * For example:
 *
 * const spy1 = Spy.on(obj, 'methodName1');
 * const spy2 = Spy.on(obj, 'methodName2');
 * const spy3 = Spy.on(obj, 'methodName3');
 *
 * Can be accomplished by:
 *
 * const obj$Mock = Spy.mock(obj, 'methodName1', 'methodName2', 'methodName3')
 *
 * (spy1 === obj$Mock.methodName1 and so forth)
 *
 * @param {Object} obj -> The manipulated object. Actual type:
 *                        Before initialization: { [$Keys<typeof methodNames>]: Throwing function }
 *                        After initialization: { [$Keys<typeof methodNames>]: SpyInstance }
 * @param {string[]} methodNames -> Iterative provided attribute
 *                                  names that will be mocked.
 *
 * @return {Object} Mock.
 */
Spy.mock = (obj, ...methodNames) => createMock(obj, methodNames);
/**
 * This static method enables you to create mocks on module scope.
 * As long as jest will support this behavior, the Spy will too. If
 * you are calling this function on other test runners you will
 * encounter an exception. You should favor to use "Spy.mock" but there
 * might be reasons that this will not work. E.g. if you want to mock
 * directly exported functions.
 *
 * For example:
 *
 * const Mock$MyModule = Spy.mockModule('./my-module', 'useMe');
 *
 * Now you could do:
 * Mock$MyModule.useMe.returns(['foo', 'bar']);
 *
 * @param {string} moduleName -> Everything that's expected by "jest.mock"
 *                               as first parameter. Relative and absolute
 *                               module paths.
 * @param {string[]} methodNames -> Iterative provided attribute
 *                                  names that will be mocked.
 *
 * @return {Object} Mock.
 */
Spy.mockModule = (moduleName, ...methodNames) => TestSuite.createModuleMock(moduleName, methodNames);
/**
 * This static method is very similar to "Spy.mockModule" but perfectly
 * suited for testing with React components. When using testing tools
 * that render the whole subtree it is sometimes better to mock parts
 * of your nested components.
 *
 * The only difference is that the returned mocks will return "null"
 * instead of "undefined" as default. This is a minimal valid React
 * component.
 *
 * @param {string} moduleName -> Everything that's expected by "jest.mock"
 *                               as first parameter. Relative and absolute
 *                               module paths.
 * @param {string[]} methodNames -> Iterative provided attribute
 *                                  names that will be mocked.
 *
 * @return {Object} Mock.
 */
Spy.mockReactComponents = (moduleName, ...methodNames) => TestSuite.createModuleMock(moduleName, methodNames, null);
/**
 * This static method initializes all created
 * mocks (see Spy.mock). This is necessary, because
 * it has to apply before each test run, to ensure
 * that restored spies apply again. This makes
 * automated cleaned up spies possible.
 *
 * Usually it should get called within one "beforeEach"-Hook.
 *
 * @param {string | void} scope -> A string identifying the scope.
 *                                 Scopes should not be used only in
 *                                 combination of custom beforeEach and
 *                                 afterEach-Hooks.
 *
 */
Spy.initMocks = (scope) => {
    initMocks(Spy.on, scope);
};
/**
 * This static method does restore all
 * manipulated objects and remove therefore
 * all mocks.
 *
 * Restoring objects does not disable any
 * other behaviors/features of the spies.
 *
 * Usually it should get called within one "afterEach"-Hook.
 */
Spy.restoreAll = () => {
    registry.restoreAll();
};
/**
 * This static method does reset all
 * created spy instances.
 *
 * This deletes all information related to made calls.
 * This is very useful, if you want to avoid testing any
 * conditions that were outside the control of your test.
 *
 * Usually it should get called within one "afterEach"-Hook.
 */
Spy.resetAll = () => {
    AllCreatedSpies.forEach((spy) => spy.reset());
};
const defaultHooks = {
    beforeEach: Spy.initMocks,
    afterEach: () => {
        Spy.restoreAll();
        Spy.resetAll();
    },
};
TestSuite.configure(defaultHooks);

export { Spy };
